#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <glib.h>

#include <bmpx/metadata.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>

typedef struct _LCMFS LCMFS;

struct _LCMFS {
    int		   i_size;
    GHashTable	  *i;

    gchar	  *index;
    gchar	  *records;

    FILE	  *f_index; 
    FILE	  *f_records; 
};

static LCMFS *lcmfs;

void
lcmfs_init ()
{
  lcmfs = g_new0 (LCMFS,1);
  
  lcmfs->i_size = 
      0;
  lcmfs->i = 
      g_hash_table_new (g_str_hash, g_str_equal);
}

void
lcmfs_open (const gchar *lcmfs_base)
{
    lcmfs->index = 
	g_strconcat (lcmfs_base, ".lif", NULL); // LCMFS Index  File
    lcmfs->records = 
	g_strconcat (lcmfs_base, ".lrd", NULL); // LCMFS Record Data

    // Open Index
    lcmfs->f_index = fopen (lcmfs->index, "r");
    if (lcmfs->f_index)
      {
	  // Read index
	  gint n;

	  fread (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
	  for (n = 0; n < lcmfs->i_size; n++)
	    {
	      uint16_t key_size;
	      gint     offset;
	      gchar   *pkey;

	      fread (&key_size, sizeof(uint16_t), 1, lcmfs->f_index);
	      pkey = g_malloc0 (key_size+1);
	      fread (pkey, sizeof(char), key_size, lcmfs->f_index);
	      fread (&offset, sizeof(gint), 1, lcmfs->f_index);

	      g_hash_table_insert (lcmfs->i, pkey, GINT_TO_POINTER (offset)); 
	    }

	  fclose (lcmfs->f_index);

      } 

    // Open Records
    lcmfs->f_records = fopen (lcmfs->records, "a+");
    // FIXME: Check errno after lockf ()
}

gboolean
lcmfs_record_exists (const gchar *pkey)
{
  return g_hash_table_lookup (lcmfs->i, pkey) ? TRUE : FALSE;
}

static void
_record_read (HTuple *record)
{
  uint8_t      value, 
	       n;
  uint16_t     value16;

  gint	       aux_int;
  gchar	      *aux_str;

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
	fread (&value, sizeof(uint8_t), 1, lcmfs->f_records);

	fread (&value16, sizeof(uint16_t), 1, lcmfs->f_records);
	aux_str = g_malloc0(value16+1);
	fread (aux_str, sizeof(char), value16, lcmfs->f_records);

	if (bmp_metadata_get_type (value) == G_TYPE_INT)
	  {
	    aux_int = strtol (aux_str, NULL, 10);
	    h_tuple_insert (record,
			    bmp_metadata_get_id_static (value),
			    G_TYPE_INT,
			    GINT_TO_POINTER(aux_int));
	  }
	else
	  {
	    h_tuple_insert (record,
			    bmp_metadata_get_id_static (value),
			    G_TYPE_STRING,
			    aux_str); 
	  }

	g_free (aux_str);
    }
}

HTuple*
lcmfs_record_get (const gchar *pkey) 
{
  HTuple      *record;
  gpointer     offset_p;
  gint	       offset;

  // Does the record exist 
  if (!lcmfs_record_exists (pkey))
      return NULL;

  record = 
      bmp_metadata_tuple_create_default ();
  offset_p =
      g_hash_table_lookup (lcmfs->i, pkey);
  g_assert (offset_p);
  offset =
      GPOINTER_TO_INT (offset_p);
  fseek (lcmfs->f_records, offset, SEEK_SET);

  _record_read (record);

  return record;
}

void
lcmfs_read_all (BmpMetadataCache *cache)
{
  gint n;

  fseek (lcmfs->f_records, 0L, SEEK_SET);

  for (n = 0; n < lcmfs->i_size; n++)
    {
      HTuple *record;

      record =
	  bmp_metadata_tuple_create_default ();

      _record_read (record);
      _bmp_metadata_cache_tuple_inserted (cache, record); 
    }
}

gboolean
lcmfs_record_add (HTuple *record)
{
  HTupleElem *elem_pkey, *elem;

  uint8_t   n; 

  uint16_t  value16;

  int	    new_offset;

  gchar	   *aux;

  // Just for the case we check if the record exists
  elem_pkey = h_tuple_get (record, (gpointer)bmp_metadata_get_id_static (BMP_METADATA_PKEY));
  if (lcmfs_record_exists (elem_pkey->value))
      return FALSE;

  // Seek to the end of the records file
  fseek (lcmfs->f_records, 0L, SEEK_END);
  new_offset = ftell (lcmfs->f_records);

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
      elem = h_tuple_get (record, bmp_metadata_get_id_static (n)); 

      fwrite (&n, sizeof(uint8_t), 1, lcmfs->f_records);
  
      if (elem->type == G_TYPE_INT)
	aux = g_strdup_printf("%d", GPOINTER_TO_INT(elem->value));
      else
	aux = g_strdup (elem->value);

      value16 = strlen(aux);
      fwrite (&value16, sizeof(uint16_t), 1, lcmfs->f_records);
      fwrite (aux, sizeof(char), strlen(aux), lcmfs->f_records);

      g_free (aux);  
    }

  g_hash_table_insert (lcmfs->i, g_strdup (elem_pkey->value), GINT_TO_POINTER (new_offset));
  lcmfs->i_size++;

  return TRUE;
}

static void
_write_index_record (gpointer key, gpointer value, gpointer data)
{
    uint16_t  key_size;
    gchar    *pkey   = (gchar*)key;
    gint      offset = GPOINTER_TO_INT (value);
 
    key_size = strlen (pkey);
 
    fwrite (&key_size, sizeof(uint16_t), 1, lcmfs->f_index); 
    fwrite (pkey, sizeof(char), key_size, lcmfs->f_index);
    fwrite (&offset, sizeof(gint), 1, lcmfs->f_index);
}

void
lcmfs_close ()
{
    lcmfs->f_index = fopen (lcmfs->index, "w");
    if (lcmfs->f_index)
      {
	  // Write index
	  fwrite (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
	  g_hash_table_foreach (lcmfs->i, (GHFunc)_write_index_record, NULL);
	  fclose (lcmfs->f_index);
      } 
    else
      {
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to write index file", G_STRLOC);
	  // ERROR
      }

    fclose (lcmfs->f_records);
}
